/*****************************************************************************
SPADE

Spherical particle detection from electron micrographs

Copyright (C )1999 T. Kaituri, E. Kallioniemi, J. Paaso,
M. Ters and J. Virtanen

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

*****************************************************************************/

/*
* CASE - Software engineering project
* University of Helsinki, Department of Computer Science
* Spring 1999
*
*
* File:     image.h
* Purpose:  Define interface for the Image class hierarchy
*
* Use:      This header is used by all the subclasses which include
*           actual implementation.
*
*
* Synopsis: Image-class hierarcy defines a Image-object type which
*           is used as container for different image types. All these
*           images have same interface, which can be used through
*           virtual Image pointer. Actual implementation is strongly
*           dependable from datatype used. Implementation is placed in
*           derived subclasses.
*
*           Mask - class is utility class used in several image
*           operation.
*           For further discussion of the operations and general class
*           structure; refer CASE-project documentation.
*
*
* Author:   Esa Kallioniemi
* Date:     15.5.1999
*
*/


#ifndef IMAGE_H_
#define IMAGE_H_

#include "mask.h"

// Datatypes defined; byte : unsigned char (8 bits)
// word : unsigned word (16 bits)
// general: unsigned char (8 bits)

enum imageType{byte, word, general};

class Image {

    public:


        virtual ~Image() {}

        //General operations

        // Image x size
        virtual int xSize() = 0;

        // Image y size
        virtual int ySize() = 0;

        // Imagetype
        virtual imageType type() = 0;

        // Physical size in bytes (8 bits / byte)
        virtual int physSize() = 0;

        // Pointer to the actual imagedata
        virtual void* physData() = 0;

        // Make copy from physical data, return it in 1D table
        virtual void* copyData() = 0;

        // Standard deviation for image
        virtual int     standardDeviation() = 0;

        // Average intensity for whole image
        virtual int     average() = 0;

        // Graphical presentation for intensity values in given
        // horizontal line
        virtual Image*  lineProfile(int xstart, int ystart, int length,
        							 int heigth) = 0;

        // Graphical precentation for histogram distribution
        virtual Image*  histogramProfile(int size) = 0;

        // Intensity in given coordinates
        virtual int     valueAt(int x, int y) = 0;

        // **************** Visual operations ************************
        
        // Not implemented
        virtual Image*  powerSpectrum2D() = 0;

        // Returns rotational average in new Image object
        virtual Image*  rotationalAverage(float step) = 0;
        

        // **************** Transformations **************************

        // Good quality scale, returns new image
        virtual Image*  scale(int destXSize, int destYSize) = 0;

        // Pad image to given size, set new area zero
        virtual int     doPadding(int xSize, int ySize) = 0;

        // Subset of the original data in a form of new Image-object
        virtual Image* 	cutArea(int upperLeftX, int upperLeftY, int xDim, 
        						int yDim) = 0;
        						
        // Histogram equalization based on cumulative histogram
        virtual Image*  equalize() = 0;
	        
        //******************* Process image ***************************

        // Definition for Mask-object used in image operations
        virtual void    setMask(Mask* msk) = 0;

        // Histogram centering
        virtual int     centerHistogram() = 0;

        // Histogram strech in standard deviations multiple for
        // left,right edge
        virtual int     transformHistogramStatisticalCut(int stdDevMul) = 0;

        // Histogram strech in percents for left,right edge
        virtual int     transformHistogramPercentCut(int lprc, int rprc) = 0;

        // Operation for blending image with defined mask.
        virtual int     blendWithMask() = 0;

        //  Optimized version of the (centerHistogram,
        //  transformHistogramStatisticalCut, blendWithMask) - operation
        //  bundle. Order is fixed.
        virtual int     compositeTransform(int stdDevMul) = 0;

        //  Optimized version of the (centerHistogram,
        //  transformHistogramPercentCut, blendWithMask) - operation
        //  bundle. Order is fixed.
        virtual int     compositeTransform(int lprc, int rprc) = 0;
        
    protected:

        // Utility methods

        virtual void* calculateHistogram()= 0;

        // Class data

        int width;              // Image x size
        int height;             // Image y size
        int averageIntensity;   // Average intensity for image
        int histogramLength;    // Lenght of histogram, equals to
                                // MAX_SIZE(datatype_in_use)
        imageType dataType;     // Objects datatype

        Mask* maskPointer;      // Pointer to the utility Mask-object
        
       	friend class Mask;

};
#endif
