/*****************************************************************************
SPADE

Spherical particle detection from electron micrographs

Copyright (C) 1999 T. Kaituri, E. Kallioniemi, J. Paaso,
M. Ters and J. Virtanen

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

*****************************************************************************/

/*
* CASE - Software engineering project
* University of Helsinki, Department of Computer Science
* Spring 1999
*
*
* File:     image8bi.cpp
* Purpose:  Implement Image8Bit-class
* Use:      This file is compiled and linked with rest of the program.
*           NOTE : use of compiler optimisations is strongly suggested.
*
* Synopsis: Image8Bit -class is implementation for Image-class object
*           using byte ( unsigned char)  data. Allmost all of the
*           operations from interface defining Image-class are given.
*           Only 2D fourier visualisation is not included in this
*           version.
*
*           Utility class Mask is used to define areas for some of the
*           operations.
*
*
* Author:   Esa Kallioniemi
* Date:     15.5.1999
*
*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "image8bi.h"
#include "globals.h"

    // ******************* public methods ******************************

/*
* Method:   Image8Bit
* Purpose:  Implement constructor
* Returns:  Nothing
* Synopsis: All the class data is initialized.
*
*/
    Image8Bit::Image8Bit(int xsize, int ysize, unsigned char* data,
                  int maxIntensity) {

        dataType = byte;        // enumeration

        width = xsize; height = ysize;
        histogramLength = maxIntensity;
        averageIntensity = -1;

        imageData = data; histogram = NULL;
        maskPointer = NULL;
    }

// Return image x size.
    int Image8Bit::xSize() { return width; }

// Return image y size.
    int Image8Bit::ySize() { return height; }

// Return image type.
    imageType Image8Bit::type() { return dataType; }

/*
* Method:   physSize
* Purpose:  Return physical size of the image data in bytes.
* Returns:  Image size in bytes (unsigned char, 8 bits)
* Synopsis:
*
*/
    int Image8Bit::physSize()  { return (width*height*sizeof(unsigned char));}

/*
* Method:   copyData
* Purpose:  Return image data in plain 1D table.
* Returns:  As above.
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Reserve memory for the data, copy original data to the
*           reserved 1D table and return data.
*
*/
    void* Image8Bit::copyData() {

		unsigned char *dataCopy;
		try {
			dataCopy = new unsigned char[this->width*this->height];
		} catch (...) {
			throw OutOfMemory(); 
		}
       	memcpy(dataCopy, this->imageData, this->physSize());
        return dataCopy;
	}

// Return pointer to physical image data.
    void* Image8Bit::physData() { return imageData; }

// Return image pixel intensity in given coordinates.
// No clipping is performed.

    int Image8Bit::valueAt(int x, int y) {
        return (int) imageData[y*width+x];
    }

/*
* Method:   standardDeviation
* Purpose:  Calculate standard deviation for image.
* Returns:  Signed int.
* Synopsis: Call utility method calculateHistogram, calculate standard
*           deviation using the new histogram.
*
*/
    int Image8Bit::standardDeviation() {

        this->calculateHistogram();

        int sum = 0,averageIntensity = 0;
        int distance = 0, count = 0;

        for(int i=0;i<histogramLength;++i) {
            sum     += i * histogram[i];
            count   += histogram[i];
        }

        averageIntensity = sum / (count+1);
        sum = 0;
        for(int i=0;i<histogramLength;++i) {
            distance = (i - averageIntensity);
            sum += ( distance * distance * histogram[i]);
        }
        return ( (int) sqrt( ( (double)sum / count) ));

    }

//  Calculate average intensity for the image. All the pixels effect the
//  result.

	int Image8Bit::average() {

        int pixelCount;
		int sum  = 0;
        unsigned char *rawData = (unsigned char *)this->physData();

        pixelCount = this->xSize() * this->ySize();
        for(int i=0;i<pixelCount;++i)
      		sum+=rawData[i];  
		
		if(pixelCount == 0)
			return 0;
		else
			return ( sum / pixelCount );
		
	}

/*
* Method:   lineProfile
* Purpose:  Return graphical representation for the single horizontal
*           scanline.
* Returns:  New Image8Bit object.
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Image data is scanned across the given scanline, these
*           scanlines intensities are represented in bars. Bar height is
*           defined by intensity. Composed image is returned in form of new
*           Image8Bit object. Last parameter is used to scale the
*           result image in given height.
*
*
*/
    Image* Image8Bit::lineProfile(int xstart, int ystart, int length, 
								   int height) {
 		double scaleFactor = 0.0;  
        unsigned char *rawData;     // Pointer to original image data
        int pixelIntensity = 0;
        int spanStart = 0;
        int bottomRow = 0;

        // Data for new image
		unsigned char *data;
         try { 
			data = new unsigned char[length*height];
		} catch (...) {
			throw OutOfMemory(); 
		}
        memset(data,255,length*height);
        rawData = (unsigned char *)this->physData();

        // Set scale
		scaleFactor = histogramLength / (double) height;
        if(scaleFactor == 0.0 ) scaleFactor = 1.0;

        // Do the drawing, one vertical line at the time
        spanStart = this->xSize() * ystart + xstart;
        bottomRow = (height-1) * length;
        for(int x=0;x<length;++x) {
            pixelIntensity = (int) rawData[spanStart+x];
 			pixelIntensity = (int) ( (double)pixelIntensity / scaleFactor); 
            for(int y=0;y<pixelIntensity;++y) {
                data[bottomRow-(y*length) + x] = 0;
            }
        }
		try {		
        	return new Image8Bit(length, height, data, 256);
		} catch (...) {
			throw OutOfMemory(); 
		}
    }
/*
* Method:   histogramProfile
* Purpose:  Return graphical representation for the images histogram
*           distribution.
* Returns:  New Image8Bit object.
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Calculate histogram using calculateHistogram method, draw
*           histogram using standard bars representation.
*           Parameter is used to control the result image height.
*
*/
    Image*  Image8Bit::histogramProfile(int size) {

        unsigned int maxHeigth = 0;
        double scaleFactor = 0.0;
        int barHeigth = 0;                                   
        this->calculateHistogram();
        int barStart = 0;

        // Data for new image
		unsigned char *data;
		try {
	  		data = new unsigned char[histogramLength*size];
		} catch (...) {
	  		throw OutOfMemory(); 
		}
        memset(data,255,size*histogramLength);

        // Find maximum height
        for(int x=0;x<histogramLength;++x) {
            if(histogram[x] > maxHeigth)
                maxHeigth = histogram[x];
        }
        // Set scale
        scaleFactor = maxHeigth / (double)size;
        if(scaleFactor == 0.0 ) scaleFactor = 1.0;

        // Do the drawing, one vertical line (bar) at the time
        barStart = (size-1) * histogramLength;
        for(int x=0;x<histogramLength;++x) {
            barHeigth = (int) ( (double)histogram[x] / scaleFactor);
            for(int y=0;y<barHeigth;++y) {
                data[barStart-(y*histogramLength) + x] = 0;
            }
        }
		try {
	  		return new Image8Bit(histogramLength, size, data,256);
		} catch (...) {
	  		throw OutOfMemory(); 
		}
    }

// NOT implemented
    Image* Image8Bit::powerSpectrum2D() {

        return NULL;
    }

/*
* Method:   rotationalAverage
* Purpose:  Calculate rotational average for the image
* Returns:  New Image8Bit object
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Image is scanned through set of angles. Angle spacing is
*           defined by given step for stepping the through full 360 degrees.
*           Scans are used to calculate average scanline which is
*           returned using graphical representation. Graphical
*           representation is bar diagram, using one bar per pixel.
*
*
*/
    Image*  Image8Bit::rotationalAverage(float step) {

        unsigned char *rawData;
        unsigned char pixelIntensity = 0;
        int bottomRow = 0;
        int length = this->xSize() / 2;
        double xGradient,yGradient;
        double angleToRad = PI / 180.0;
        double origoX = (double)(width / 2);
        double origoY = (double)(height / 2);
        int rotations = (int)(360.0 / step);
        int rotStep = (int) step;

        // Data for new image
		unsigned char *data;
		try {
        	data = new unsigned char[length*histogramLength]; 
		} catch(...) {
	  		throw OutOfMemory();
		}

        memset(data,255,length*histogramLength);

        // Table for average scanline
        int *profile; 
		try {
			profile = new int[length];
		} catch (...) {  
	 		throw OutOfMemory();
		}
        memset(profile,255,length*sizeof(int));

        rawData = (unsigned char *)this->physData();

        // Step throught image one angle at the time
        for(int i=0;i<360;i+=rotStep) {
            xGradient = cos(angleToRad*i);
            yGradient = sin(angleToRad*i);

            origoX = (double)(width / 2);
            origoY = (double)(height / 2);
  
            for(int j=0;j<length;++j){
                profile[j] += rawData[width*(int)origoY+(int)origoX];
                origoX+=xGradient;
                origoY+=yGradient;
            }
        }

        // Calculate average for each scanline element
        for(int j=0;j<length;++j) {
            profile[j] /= rotations;
        }

        // Draw image in vertical lines
        // Pixel intensity defines bar height
        bottomRow = (histogramLength-1) * length;
        for(int x=0;x<length;++x) {
            pixelIntensity = (unsigned char) profile[x];
            for(int y=0;y<pixelIntensity;++y) {
                data[bottomRow-(y*length) + x] = 0;
            }
        }
		try {
        	return new Image8Bit(length, histogramLength, data, 256);
		} catch (...) {
	  		throw OutOfMemory();
		}

    }
    

// *********************** Transformations **************************

/*
* Method:   scale
* Purpose:  Return up or down scaled version of the image
* Returns:  New Image8Bit object
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Scale image using averaging (box filter) and bi-linear
*           per pixel filtering when downscaling. Use plain integer
*           nearest pixel sampling for upscaling.
*
*
*/
    Image*  Image8Bit::scale(int destXSize, int destYSize) {

        unsigned char *destData;
		try {
			destData = new unsigned char[destXSize*destYSize];
		} catch (...) {
	  		throw OutOfMemory();
		}

        unsigned int  *spanBuffer;
        unsigned char *sourceData = (unsigned char *)this->physData();
        unsigned char *averageData;
        unsigned char *destPointer = destData;
        unsigned char *sourcePointer = sourceData;
        unsigned char *upperLine;
        unsigned char *lowerLine;

        int xPixel,yPixel;
        int averageX,averageY;
        bool averageIsDone = false;
        int yBlockSize,xBlockSize,pointerInBlock, areaSize;

        double scaleRatioX = (double)this->xSize() / double(destXSize);
        double scaleRatioY = (double)this->ySize() / double(destYSize);
        double U = 0.0,V = 0.0;
        double fracU = 0.0,fracV = 0.0,invFracU = 0.0,invFracV = 0.0;

        // Destination == Source -> return plain copy
        if( ( scaleRatioX == 1.0) && (scaleRatioY == 1.0) ) {
            memcpy(destData, this->imageData, this->physSize());
	    	try {
            	return new Image8Bit(destXSize, destYSize, destData, 256);
	    	} catch (...) {
	      		throw OutOfMemory();
	    	}
        }

        // Test if we are sampling upwards
        if( ( scaleRatioX < 1.0) || (scaleRatioY < 1.0) ) {

            unsigned int xPos = 0;
            unsigned int yPos = 0;
            unsigned int xStep = (this->xSize() << 16) / destXSize;
            unsigned int yStep = (this->ySize() << 16) / destYSize;

            // Sample using nearest pixel approach
            // Use fixed point math for gradients
            for(int y=0;y<destYSize;++y) {
                xPos = 0;
                for(int x=0;x<destXSize;++x) {
                    *destPointer = sourceData[xPos >> 16];

                    xPos+=xStep;    // Step x gradient
                    ++destPointer;  // Step destination
                }
                // Step source y scanlines
                yPos+=yStep;
                sourceData =  (unsigned char *)this->physData();
                sourceData += (yPos >> 16)*this->xSize();
            }
			try {
            return new Image8Bit(destXSize, destYSize, destData, 256);
			} catch(...) {
				throw OutOfMemory();
			}
        }

        // Test if full transform is needed
        if( (2.0 <= scaleRatioX ) || (2.0 <= scaleRatioY ) ) {
            averageIsDone = true;

            // Get block size for averaging
            xBlockSize = (int) scaleRatioX;
            yBlockSize = (int) scaleRatioY;

            if(xBlockSize == 0) xBlockSize = 1;
            averageX = this->xSize() / xBlockSize;

            if(yBlockSize == 0) yBlockSize = 1;
            averageY = this->ySize() / yBlockSize;

            averageData = new unsigned char[averageX*averageY];
            unsigned char *writePointer = averageData;
            spanBuffer  = new unsigned int [averageX];
            areaSize = xBlockSize * yBlockSize;

            // Cumulate all the source x spans needed for single
            // destination x span
            for(int y=0;y<averageY;++y) {
            // Spans are cumulated to spanBuffer
                memset(spanBuffer,0,averageX * sizeof(unsigned int));

                for(int lines=0;lines<yBlockSize;++lines) {
                    for(int x=0;x<averageX;++x) {
                        pointerInBlock = 0;
                        while(pointerInBlock < xBlockSize) {
                            spanBuffer[x] += *sourcePointer;
                            ++sourcePointer;
                            ++pointerInBlock;
                        }
                    }
                }
                // Calculate cumulated spans average, and write it out
                for(int x=0;x<averageX;++x)
                    writePointer[x] = (unsigned char) (spanBuffer[x] / areaSize);
                writePointer +=averageX;    // Move one line down
            }
            scaleRatioX = (double)averageX / double(destXSize);
            scaleRatioY = (double)averageY / double(destYSize);
        }
        else {  // No average is needed
            averageData = sourceData;
            averageX = width;
        }

        // Do the bi-linear filtering
        // for each pixel
            for(int y=0;y<destYSize;++y) {
                yPixel      = (int) V;
                fracV       = V - yPixel;
                invFracV    = 1.0 - fracV;
                upperLine   = averageData+(yPixel * averageX);
                lowerLine   = averageData+((yPixel+1) * averageX);
                U = 0.0;

                for(int x=0;x<destXSize;++x) {
                    xPixel      = (int) U;
                    fracU       = U - xPixel;
                    invFracU    = 1.0 - fracU;

                    *destPointer= (unsigned char)
                                  ( invFracV *
                                  ( (invFracU * upperLine[xPixel])
                                  + (fracU * upperLine[xPixel+1]) )
                                  + fracV *
                                  ( (invFracU * lowerLine[xPixel])
                                   +(fracU * lowerLine[xPixel+1])) );
                    ++destPointer;
                    U +=scaleRatioX;
                }
                V +=scaleRatioY;
            }

        if(averageIsDone) delete [] averageData;
		try {
        	return new Image8Bit(destXSize, destYSize, destData, 256);
		} catch (...) {
			throw OutOfMemory();
		}
    }


/*
* Method:   doPadding
* Purpose:  Enlarge image, keep old data as subset of new image
* Returns:  -1 for error, 1 otherwise
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Pad image to new dimensions, only upward enlargement is
*           possible. Old image data is placed to the upperleft corner.
*
*/
    int     Image8Bit::doPadding(int xSize, int ySize) {

        // Only upwards !!
        if( (xSize < this->width) || (ySize < this->height) )
            return -1;

        // Space for enlarged image
        unsigned char *destData; 
		try {
			destData = new unsigned char[xSize*ySize];
		} catch (...) {
			throw OutOfMemory();
		}

        memset(destData,255,xSize*ySize);

		unsigned char *sourceData = (unsigned char *)this->physData();
        unsigned char *destPointer = destData; 

        int xDim = this->width;     // source x size
        int yDim = this->height;    // source y size

        // Copy old image data to the upperleft corner
        for(int y=0;y<yDim;++y) {
            memcpy(destData,(sourceData+(y*width)),xDim*sizeof(unsigned char));
            destData +=xSize;
		}		

        // Replace old image data, old size attributes
        delete [] this->physData();
        imageData = destPointer;
        width = xSize; height = ySize;
        return 1;
    }

/*
* Method:   cutArea
* Purpose:  Return subset of the original image data
* Returns:  New Image8Bit object
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Cut area from image data using giver parameters, return data
*           in Image8Bit object.
*
*/
    Image * Image8Bit::cutArea(int upperLeftX, int upperLeftY, int xDim,
							   int yDim) {

		int destXSize = xDim;
		int destYSize = yDim;

		unsigned char *destData; 
		try {
			destData= new unsigned char[destXSize*destYSize];
		} catch (...) {
			throw OutOfMemory(); 
		}

		unsigned char *sourceData = (unsigned char *)this->physData();
        unsigned char *destPointer = destData; 
		sourceData += (this->width * upperLeftY + upperLeftX);
		 
		for(int y=0;y<destYSize;++y) {
			memcpy(destData,(sourceData+(y*width)),xDim*sizeof(unsigned char));
			destData +=xDim;
		}		
		try {
        	return new Image8Bit(destXSize, destYSize, destPointer, 256);
		} catch (...) {
			throw OutOfMemory(); 
		}
	} 

/*
* Method:   equalize
* Purpose:  Equalize image
* Returns:  New Image8Bit object
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Copy image data to the buffer, calculate cumulative
*           histogram and equalized data using this histogram. Form new
*           Image8Bit object using data, return it.
*
*/
	Image * Image8Bit::equalize() { 

        int pixelCount;		
		float scale;

        // Space for image data copy
        unsigned char *pic;
		try { 
			pic = new unsigned char[this->width*this->height];
		} catch (...) {
			throw OutOfMemory(); 
		}
        
        memcpy(pic, this->imageData, this->physSize());

      	unsigned int *h;
		try {
			h = new unsigned int[this->histogramLength];
		} catch (...) {
			throw OutOfMemory(); 
		}
		 
		memset(h,0,this->histogramLength*sizeof(unsigned int));

        // Calculate histogram
        pixelCount = this->xSize() * this->ySize();
        for(int i=0;i<pixelCount;++i)
            h[ pic[i] ]++;

        // Calculate cumulation
		for (int i = 1; i <histogramLength; i++) 
        	h[i] += h[i - 1];
   			                             
        scale = ((float) histogramLength-1.0) / ((float) h[histogramLength-1]);

        // Scale histogram
        for (int i = 0; i <histogramLength; i++)
        	h[i] = (int) (h[i] * scale);

        // Transform data
		for(int i=0;i<pixelCount;++i)
			pic[i]= h[ pic[i] ];  
		
		delete [] h;
		try {
			return new Image8Bit(this->xSize(), this->ySize(), pic, 256);
		} catch (...) {
			throw OutOfMemory(); 
		}
		
	}

// ********************* Image processing *****************************

// Set pointer to the Mask-object used in image operations.
    void    Image8Bit::setMask(Mask* msk) { maskPointer = msk;}

/*
* Method:   centerHistogram
* Purpose:  Tranform image using histogram centering
* Returns:  0 in error, 1 otherwise
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Using predefined mask scan area outside inner circle and
*           calculate average intensity for this area. Calculate
*           diffrence needed to move intensities so that average
*           intensity equals histogram center. Modify image pixel
*           intensities using calculated offset table.
*
*/
    int     Image8Bit::centerHistogram() {

        if(maskPointer == NULL) return 0;

        // Offset table
        unsigned char *cTable;
 		try {
			cTable = new unsigned char[this->histogramLength];
		} catch (...) {
			throw OutOfMemory(); 
		}

        unsigned char *image= (unsigned char *)this->physData();
        unsigned char *spanStart;
        Mask *mp = maskPointer;

        int intensitySum    = 0;
        int top             = height / 2 - mp->innerRad;
        int bottom          = height / 2 + mp->innerRad;

        int begin = 0,end = 0;
        int x,y,pixelCount = 1;
        int intensity = 0;
        int xOffset = 0;

        // Scan area outside inner circle, calculate average
        for(y=0;y<this->height;++y) {

            if( (y >= top) && (y <= bottom) ) {
                spanStart = image + ( width * y);
                begin = 0;
                end = mp->innerSpanLeft[y];
                pixelCount+=end;
                for(x=begin;x<end;++x) {
                    intensitySum += spanStart[x];
                }

                // We discard inner circle area.
                
                pixelCount+= width - mp->innerSpanRight[y];
                begin = mp->innerSpanRight[y];
                end   = width;
                for(x=begin;x<end;++x) {
                    intensitySum += spanStart[x];
                }
            }
            else {
                spanStart = image + ( width * y);
                for(x=0;x<width;++x) {
                    intensitySum += spanStart[x];
                }
                pixelCount+=width;
            }

        }

        // Calculate average intensity
        averageIntensity = intensitySum / pixelCount;

        // Calculate offset between histogram center and average
        // intensity
        xOffset = (histogramLength / 2 ) - averageIntensity;

        // Build offset table
        if(xOffset < 0)     // We move intensity values left
            for(x=0;x<histogramLength;++x) {
                intensity = x + xOffset;
                if(intensity < 0) intensity = 0;
                cTable[x] = (unsigned char) intensity;
            }
        else {              // We move intensity values right
            for(x=0;x<histogramLength;++x) {
                intensity = x + xOffset;
                if(intensity > histogramLength) intensity = histogramLength;
                cTable[x] = (unsigned char) intensity;
            }
        }

        // Transform whole image using offset table
        pixelCount = height * width;
        for(int i=0;i<pixelCount;++i)
            image[i] = cTable[image[i]];

        delete [] cTable;
        return 1;
    }

/*
* Method:   transfromHistogramStatisticalCut
* Purpose:  Perform contrast strech for the image
* Returns:  1 if ok
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Modify image using contrast strech, perform strech using
*           linear interpolation for sampling. Define strech left and
*           right endpoints in histogram using multiple of standard
*           deviation. Strech defined area so that it covers whole
*           histogram range.
*
*/
    int     Image8Bit::transformHistogramStatisticalCut(int stdDevMul) {

        // Get histogram for current image data
        this->calculateHistogram();

        int begin,end,pixelCount;
        int sum = 0,averageIntensity = 0;
        int standardDeviation = 0;
        int distance = 0, count = 0;

        // Offset table, used as substitute for histogram
        unsigned char *cTable;
        try {
            cTable = new unsigned char[this->histogramLength];
        } catch (...) {
            throw OutOfMemory();
        }  
        memset(cTable,0,this->histogramLength*sizeof(unsigned char));

        unsigned char *image= (unsigned char *)this->physData();

        // Calculate average
        for(int i=0;i<histogramLength;++i) {
            sum     += i * histogram[i];
            count   += histogram[i];
        }
        averageIntensity = sum / (count+1);
        sum = 0;

        // Calculate standard deviation
        for(int i=0;i<histogramLength;++i) {
            distance = (i - averageIntensity);
            sum += ( distance * distance * histogram[i]);
        }
        standardDeviation = (int) sqrt( ( (double)sum / count) );

        // Set left and right edge for strech
        begin   = averageIntensity - (standardDeviation * stdDevMul);
        end     = averageIntensity + (standardDeviation * stdDevMul);

        // Clips edges to the histogram dimensions
        if(end >= histogramLength) end = histogramLength - 1;
        if(begin < 0) begin = 0;

        // Calculate interpolation step
        int sourceLength = end - begin;
        int mulStep = ((histogramLength - 1) << 16) /  sourceLength;

        // Interpolate
        for(int x=0;x<sourceLength+1;++x) {
            cTable[begin] = (mulStep * x) >> 16;
            ++begin;
        }
        // Set intensities above right edge to the max intensity
        for(int x=end+1;x<histogramLength;++x)
            cTable[x] = histogramLength;       // h.length == max intensity

        // Tranform whole image
        pixelCount = this->height * this->width;
        for(int i=0;i<pixelCount;++i)
            image[i] = cTable[image[i]];

        delete [] cTable;

        return 1;
    }

/*
* Method:   transfromHistogramPercentCut
* Purpose:  Perform contrast strech for the image
* Returns:  1 if ok
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Modify image using contrast strech, perform strech using
*           linear interpolation for sampling. Define strech left and
*           right endpoints in histogram using percents from histogram
*           start and end points.
*           Strech defined area so that it covers whole histogram range.
*
*/
    int     Image8Bit::transformHistogramPercentCut(int lprc, int rprc) {

        int begin,end,pixelCount;

        // Offset table, used as substitute for histogram
        unsigned char *cTable;
        try {
            cTable = new unsigned char[this->histogramLength];
        } catch (...) {
            throw OutOfMemory();
        }  
        memset(cTable,0,this->histogramLength*sizeof(unsigned char));

        unsigned char *image= (unsigned char *)this->physData();

        // Calculate left and right end points
        begin   = lprc * histogramLength / 100;
        end     = ( (100 - rprc) * histogramLength) / 100;

        // Clips edges to the histogram dimensions
        if(end >= histogramLength) end = histogramLength - 1;

        // Calculate interpolation step
        int sourceLength = end - begin;
        int mulStep = ((histogramLength - 1) << 16) /  sourceLength;

        // Interpolate
        for(int x=0;x<sourceLength+1;++x) {
            cTable[begin] = (mulStep * x) >> 16;
            ++begin;
        }

        // Set intensities above right edge to the max intensity
        for(int x=end+1;x<histogramLength;++x)
            cTable[x] = histogramLength;       // h.length == max intensity

        // Transform whole image
        pixelCount = this->height * this->width;
        for(int i=0;i<pixelCount;++i)
            image[i] = cTable[image[i]];

        delete [] cTable;

        return 1;
    }

/*
* Method:   blendWithMask
* Purpose:  Blend image with previously set mask. Mask have to be set
*           with setMask -method.
* Returns:  0 if no mask is set, 1 else.
* Synopsis: Image data is modified using mask. Area enclosed by inner
*           circle is left untouched. Area outside outer circle is set
*           to average intensity. Blend area between inner and outer
*           circle is blended with cosine fuction based alpha channel slide.
*           Blend pixels, inner and outer circle tables are read from Mask
*           object and used to modify image.
*
*/
    int     Image8Bit::blendWithMask() {

        if(maskPointer == NULL) return 0;

        unsigned char *image= (unsigned char *)this->physData();
        unsigned char *spanStart;
        Mask *mp = maskPointer;

        int top             = height / 2 - mp->outerRad;
        int bottom          = height / 2 + mp->outerRad;

        int begin = 0,end = 0;
        int x,y;
        unsigned char middleIntensity = (unsigned char)(histogramLength / 2 );
        float alpha = 0;

        // Set outside area to the average intensity
        for(y=0;y<this->height;++y) {

            if( (y >= top) && (y <= bottom) ) {
                spanStart = image + ( width * y);
                begin = 0;
                end = mp->outerSpanLeft[y];
                for(x=begin;x<end;++x) {
                    spanStart[x] = middleIntensity;
                }

                // We discard inner circle area.
                
                begin = mp->outerSpanRight[y];
                end   = width;
                for(x=begin;x<end;++x) {
                    spanStart[x] = middleIntensity;
                }
            }
            else {
                spanStart = image + ( width * y);
                for(x=0;x<width;++x) {
                    spanStart[x] = middleIntensity;
                }
            }

        }
        // Do alpha pixels. (This code blows, could do 10 times better
        //                   but don't have time :( )
        int address = 0;
        for(int i=0;i<mp->alphaPixelsAmount;++i) {
            alpha = mp->blendPixels[i].value;
            address = mp->blendPixels[i].x + (mp->blendPixels[i].y * width);
            image[address] = (unsigned char) ( (alpha * middleIntensity) +
            (1-alpha) * image[address] );
        }

        return 1;
    }

/*
* Method:   compositeTransform
* Purpose:  Implement optimised composite transform.
* Returns:  1 if ok, -1 else
* Synopsis: Optimised, fixed order transform bundle including
*           centerHistogramm transformHistogramStatisticalCut, blendWithMask
*           - operations (in that order).
*
*           NOTE : Current version doesn't include any optimisations.
*/
    int     Image8Bit::compositeTransform(int stdDevMul) {

        this->centerHistogram();
        this->transformHistogramStatisticalCut(stdDevMul);
        this->blendWithMask();

        return 1;
    }

/*
* Method:   compositeTransform
* Purpose:  Implement optimised composite transform.
* Returns:  1 if ok, -1 else
* Synopsis: Optimised, fixed order transform bundle including
*           centerHistogramm transformHistogramPercentCut, blendWithMask
*           - operations (in that order).
*
*           NOTE : Current version doesn't include any optimisations.
*/
    int     Image8Bit::compositeTransform(int lprc, int rprc) {

        this->centerHistogram();
        this->transformHistogramPercentCut(lprc,rprc);
        this->blendWithMask();

        return 1;
    }

/*
* Method:   ~Image8Bit
* Purpose:  Implement destructo for the class
* Returns:  Nothing
* Synopsis: Destructor is implemented.
*
*/
    Image8Bit::~Image8Bit() {
        if(imageData != NULL) delete [] imageData;
        if(histogram != NULL) delete [] histogram;
    }

// *********************** Protected methods ************************

/*
* Method:   calculateHistogram
* Purpose:  Calculate histogram for the image. Reserve histogram table
*           if needed.
* Returns:  Pointer to the new histogram.
*           Throws OutOfMemory exception - if memory allocation fails.
*
* Synopsis: Histogram table is reserved dynamically if needed, histogram
*           is calculated. Class variable histogram points to the new
*           histogram.
*
*
*/
    void* Image8Bit::calculateHistogram() {

        int pixelCount;
        unsigned char *rawData = (unsigned char *)this->physData();
		unsigned int *h;

        if(histogram == NULL) {
          
        	try {
           		h = new unsigned int[this->histogramLength];
        	} catch (...) {
            	throw OutOfMemory();
        	}  
            histogram = h;
        }

        memset(histogram,0,this->histogramLength*sizeof(unsigned int));
        pixelCount = this->xSize() * this->ySize();
        for(int i=0;i<pixelCount;++i)
            histogram[(rawData[i])]++;

        return histogram;

    }

