/*****************************************************************************
SPADE

Spherical particle detection from electron micrographs

Copyright (C) 1999 T. Kaituri, E. Kallioniemi, J. Paaso,
M. Ters and J. Virtanen

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

*****************************************************************************/
    
/*
* CASE - Software engineering project
* University of Helsinki, Department of Computer Science
* Spring 1999
*
*
* File:     mask.cpp
* Purpose:  Implement Mask-class
* Use:      This file is linked with modules using Mask objects.
*
*
* Synopsis: Mask - class is utility class used in Image-class hierarcy.
*           It consist of two circles, outer and inner circle. Circles
*           define areas for use of different image tranformations.
*           Circle data is build using midpoint circle algorithm.
*           Additionally Mask includes definition for blend area between
*           two circles. Blend area consist of set of pixels, each
*           having x,y coordinates and alpha channel value. Alpha
*           value is defined using cosine function which is indexed
*           using distance from inner circle to outer circle.
*
*           Mask is always constructed for fixed size image.
*
* Author:   Esa Kallioniemi
* Date:     15.5.1999
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>


#include "mask.h"
#include "globals.h"

/************************** Protypes ************************************/
    void circle(signed int *left, signed int *right,
                int xCenter, int yCenter, int radius);

/*
* Method:   mask
* Purpose:  Implement constructor
* Returns:  Nothing
*           Throws OutOfMemory excaeption - if memory allocation fails
*
* Synopsis: Generate mask, consisting of two circles and structure
*           based table for alpha pixels. Use circle utility method for
*           circle edge creation.
*
*/
    Mask::Mask(int ray, int alphaEdgeSize,int imageXSize, int imageYSize){

        xSize       = imageXSize;
        ySize       = imageYSize;
        innerRad    = ray;
        outerRad    = ray + alphaEdgeSize;
        int xCent   = xSize / 2;
        int yCent   = ySize / 2;

        int outerBoxTop     = yCent - outerRad;
        int outerBoxBottom  = yCent + outerRad;
        int innerBoxTop     = yCent - innerRad;
        int innerBoxBottom  = yCent + innerRad;
        int pos = 0;
		int size = 0;
        int begin = 0,end = 0;
        int xGrad = 0, yGrad = 0;
        double distance  = 0.0;
        double innerDistance = (double)innerRad;

        alphaPixelsAmount = 0;

		double *cosTable = new double[alphaEdgeSize+1];

		try { outerSpanLeft = new signed int[ySize]; } catch (...) {
			throw OutOfMemory();
		}

        for(int i=0;i<ySize;++i) outerSpanLeft[i] =  INT_MAX;

        try { outerSpanRight = new signed int[ySize]; } catch (...) {
			throw OutOfMemory(); 
		}
        memset(outerSpanRight,0,ySize*sizeof(signed int));

        try { innerSpanLeft = new signed int[ySize]; } catch (...) {
			throw OutOfMemory(); 
		}

        for(int i=0;i<ySize;++i) innerSpanLeft[i] = INT_MAX;

        try { innerSpanRight = new signed int[ySize]; } catch (...) {
			throw OutOfMemory(); 
		}
        memset(innerSpanRight,0,ySize*sizeof(signed int));

        // Generate circles
        circle(outerSpanLeft, outerSpanRight, (xSize / 2), (ySize / 2),
				outerRad);
        circle(innerSpanLeft, innerSpanRight, (xSize / 2), (ySize / 2),
               innerRad);

        // Generate cosine table for alpha blending
        double radianStep = (PI * 0.5) / (double)(alphaEdgeSize+1);
        for(int i=0;i<=alphaEdgeSize;i++)
            cosTable[i] = 1.0 - cos(radianStep * (double)i);

        // Scan the area between inner and outer circle to
        // calculate amount of pixels in the area
         for(int y=0;y<ySize;++y) {
            if(y >= outerBoxTop && y <= outerBoxBottom)
                if(y >= innerBoxTop && y <= innerBoxBottom) {

                    begin   = outerSpanLeft[y];
                    end     = innerSpanLeft[y];
              		size+=(end-begin);  

                    begin   = innerSpanRight[y];
                    end     = outerSpanRight[y];
                  
					 size+=(end-begin);  
                }
                else {
                    begin   = outerSpanLeft[y];
                    end     = outerSpanRight[y];
              		size+=(end-begin);
                }
        }

        // Reserve space for alpha pixels
		try { blendPixels = new struct alphaPixel[size]; } catch (...) {
			throw OutOfMemory(); 
		}
        memset(blendPixels,0,( size * sizeof(struct alphaPixel)) );

        // Scan the area between inner and outer circle
        // calculate distance from inner circle to outer circle
        // Use distance to set alpha intensity to current pixel
        // Write pixels to the alpha pixels table

        for(int y=0;y<ySize;++y) {
            if(y >= outerBoxTop && y <= outerBoxBottom)

                if(y >= innerBoxTop && y <= innerBoxBottom) {

                    begin   = outerSpanLeft[y];
                    end     = innerSpanLeft[y];

                    yGrad = yCent - y ;
                    yGrad = yGrad * yGrad;
                    for(int x=begin;x<end;++x) {
                        xGrad = xCent - x;
                        distance = sqrt( (double)( xGrad*xGrad + yGrad));
                        blendPixels[pos].x     = x;
                        blendPixels[pos].y     = y;
                        blendPixels[pos].value =
                        cosTable[(int)(floor((distance - innerDistance)))];
                        ++pos;
                    }
                    begin   = innerSpanRight[y];
                    end     = outerSpanRight[y];
                    for(int x=begin;x<end;++x) {
                        xGrad = xCent - x;
                        distance = sqrt( (double)( xGrad*xGrad + yGrad));
                        blendPixels[pos].x     = x;
                        blendPixels[pos].y     = y;
                        blendPixels[pos].value =
                            cosTable[(int)(floor((distance - innerDistance)))];
                        ++pos;
                    }
                }
                else {
                    // Only outer circle edges in this scanline
                    begin   = outerSpanLeft[y];
                    end     = outerSpanRight[y];
                    yGrad = yCent - y ;
                    yGrad = yGrad * yGrad;
                    for(int x=begin;x<end;++x) {
                        xGrad = xCent - x;
                        distance = sqrt( (double)( xGrad*xGrad + yGrad));
                       	blendPixels[pos].x     = x;
                       	blendPixels[pos].y     = y;
                       	blendPixels[pos].value =
                             cosTable[(int)(floor((distance - innerDistance)))];
                      ++pos;
						
                    }
                }
        }
        alphaPixelsAmount = pos;
        delete [] cosTable;
    }

// Following methods are NOT implemented
/*  void Mask::change(int ray, int alphaEdgeSize) {
        ray = alphaEdgeSize;
    }

    void Mask::changeInnerArea(int ray) {
        ray = ray * 1;
    }

    void Mask::changeBlendArea(int alphaEdgeSize) {
       alphaEdgeSize = 0;
    }
*/

/*
* Method:   ~mask
* Purpose:  Implement destructor
* Returns:  Nothing
* Synopsis: All the possibly reserved data deleted.
*
*
*/
    Mask::~Mask() {

        if(innerSpanLeft    != NULL ) delete [] innerSpanLeft;
        if(innerSpanRight   != NULL ) delete [] innerSpanRight;
        if(outerSpanLeft    != NULL ) delete [] outerSpanLeft;
        if(outerSpanRight   != NULL ) delete [] outerSpanRight;
		if(blendPixels		!= NULL ) delete [] blendPixels;
    }

/*********************** Utility functions ******************************/

/*
* Method:   circle
* Purpose:  Implement scanline based circle algorithm
* Returns:  Nothing
* Synopsis: Generate circles left and right edges for given tables, in
*           given size. Algorithm used is midpoint circle algorithm,
*           using only integer arithmetic.
*
*/
    void circle(signed int *left, signed int *right,
                int xCenter, int yCenter, int radius) {

        int xStep   = 0;
        int yStep   = radius;
        int p       = 1-radius;

        left[yCenter+xStep] = xCenter-yStep;

        left[yCenter+yStep] = xCenter-xStep;

        left[yCenter-yStep] = xCenter-xStep;

        left[yCenter-xStep] = xCenter-yStep;

        right[yCenter+xStep] = xCenter+yStep;

        right[yCenter+yStep] = xCenter+xStep;

        right[yCenter-yStep] = xCenter+xStep;

        right[yCenter-xStep] = xCenter+yStep;

        while(xStep < yStep) {

            ++xStep;
            if(p < 0) {
                p = p + (xStep * 2) + 1;
            }
            else {
                --yStep;
                p = p + ( (xStep - yStep) * 2) + 1;
            }

            left[yCenter+xStep] = xCenter-yStep;

            if(left[(yCenter+yStep)] > (xCenter-xStep))
                left[yCenter+yStep] = xCenter-xStep;

            if(left[(yCenter-yStep)] > (xCenter-xStep))
                left[yCenter-yStep] = xCenter-xStep;

            left[yCenter-xStep] = xCenter-yStep;

            right[yCenter+xStep] = xCenter+yStep;

            if(right[(yCenter+yStep)] < (xCenter+xStep))
                right[yCenter+yStep] = xCenter+xStep;

            if(right[(yCenter-yStep)] < (xCenter+xStep))
                right[yCenter-yStep] = xCenter+xStep;

            right[yCenter-xStep] = xCenter+yStep;

        }
    }

